import "./polyfill/promise";
import { getStyle, setStyles } from "./dom/css";
import { merge } from "./extend";
const elementAnimations = new WeakMap();
export const EASE_LINEAR = (t) => t;
export const EASE_IN_QUAD = (t) => t * t;
export const EASE_OUT_QUAD = (t) => t * (2 - t);
export const EASE_IN_OUT_QUAD = (t) => t < 0.5 ? 2 * t * t : -1 + (4 - 2 * t) * t;
export const EASE_IN_CUBIC = (t) => t * t * t;
export const EASE_OUT_CUBIC = (t) => ((--t) * t * t) + 1;
export const EASE_IN_OUT_CUBIC = (t) => t < 0.5 ? 4 * t * t * t : ((t - 1) * (2 * t - 2) * (2 * t - 2)) + 1;
export const EASE_IN_QUART = (t) => t * t * t * t;
export const EASE_OUT_QUART = (t) => 1 - (--t) * t * t * t;
export const EASE_IN_OUT_QUART = (t) => t < 0.5 ? 8 * t * t * t * t : 1 - 8 * (--t) * t * t * t;
export const EASE_IN_QUINT = (t) => t * t * t * t * t;
export const EASE_OUT_QUINT = (t) => 1 + (--t) * t * t * t * t;
export const EASE_IN_OUT_QUINT = (t) => t < 0.5 ? 16 * t * t * t * t * t : 1 + (16 * (--t) * t * t * t * t);
export function animate(element, properties, options = {}) {
    let values = null;
    if (options.stopPrevious === undefined || options.stopPrevious) {
        stopAnimation(element);
    }
    const director = animateCallback((progress) => {
        if (null === values) {
            values = fetchPropertyValues(element, properties);
        }
        else {
            applyAllInterpolatedValues(element, values, progress);
        }
    }, options);
    elementAnimations.set(element, director);
    const onDone = () => {
        elementAnimations.delete(element);
    };
    director.then(onDone, onDone);
    return director;
}
export function animateCallback(callback, options = {}) {
    const context = merge({
        duration: 400,
        easing: EASE_IN_OUT_CUBIC,
    }, options);
    context.currentFrame = null;
    const animationDirector = new Promise(resolve => {
        context.onAnimationFinished = resolve;
    });
    context.currentFrame = window.requestAnimationFrame(time => runAnimationStep(time, time, callback, context));
    animationDirector.stop = () => {
        context.onAnimationFinished(false);
        if (null !== context.currentFrame) {
            window.cancelAnimationFrame(context.currentFrame);
        }
    };
    return animationDirector;
}
function runAnimationStep(time, start, callback, context) {
    const linearProgress = Math.min(1, (time - start) / context.duration);
    const easedProgress = context.easing(linearProgress);
    callback(easedProgress);
    if (linearProgress < 1) {
        context.currentFrame = window.requestAnimationFrame(time => runAnimationStep(time, start, callback, context));
    }
    else {
        context.onAnimationFinished(true);
    }
}
export function stopAnimation(element) {
    let director = elementAnimations.get(element);
    if (director !== undefined) {
        director.stop();
        elementAnimations.delete(element);
    }
}
function fetchPropertyValues(element, properties) {
    const values = {};
    for (const property in properties) {
        let rawStart = getStyle(element, property);
        let start = null !== rawStart
            ? rawStart
            : 0;
        values[property] = {
            start: start,
            delta: properties[property] - start,
        };
    }
    return values;
}
function applyAllInterpolatedValues(element, initialValues, progress) {
    const updates = {};
    for (const property in initialValues) {
        updates[property] = initialValues[property].start + (initialValues[property].delta * progress);
    }
    setStyles(element, updates);
}
